/*----------------------------------------------------------------------
  File    : cmfilter.c
  Contents: linear filter repository for closed and maximal patterns
  Author  : Christian Borgelt
  History : 2016.10.19 file created
            2016.10.20 first version completed
            2016.10.21 function cmf_report() added (all item sets)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include "arrays.h"
#include "cmfilter.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- (frequent) item set --- */
  RSUPP supp;                   /* support of the item set */
  ITEM  size;                   /* number of items of the item set */
  ITEM  items[1];               /* array of items */
} ITEMSET;                      /* ((frequent) item set) */

typedef struct {                /* --- closed/max. filter block --- */
  void  *succ;                  /* pointer to successor block */
  void  *end;                   /* end of data stored in this block */
} CMFBLK;                       /* (closed/maximal filter block) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

CMFILTER* cmf_create  (size_t blksize)
{                               /* --- create a closed/max. filter */
  CMFILTER *cmf;                /* created a closed/maximal filter */
  CMFBLK   *blk;                /* first memory block */

  if (blksize < 4096) blksize = (size_t)(1024*1024);
  cmf = (CMFILTER*)malloc(sizeof(CMFILTER));
  if (!cmf) return NULL;        /* create the base structure */
  blk = (CMFBLK*)malloc(blksize *sizeof(char));
  if (!blk) { free(cmf); return NULL; }
  blk->succ    = NULL;          /* create first memory block */
  blk->end     = blk+1;         /* and initialize it */
  cmf->iscnt   = 0;             /* no item sets stored yet */
  cmf->blkcnt  = 1;             /* only one memory block in list */
  cmf->blksize = blksize;       /* note the (standard) block size */
  cmf->head    = blk;           /* store the first (and only) block */
  cmf->end     = ((char*)blk) +blksize;
  return cmf;                   /* return created closed/max. filter */
}  /* cmf_create() */

/*--------------------------------------------------------------------*/

void cmf_delete (CMFILTER *cmf)
{                               /* --- delete a closed/max. filter */
  CMFBLK *blk;                  /* to traverse the memory block list */

  assert(cmf);                  /* check the function argument */
  while (cmf->head) {           /* while the block list is not empty */
    blk = (CMFBLK*)cmf->head;   /* get the next memory block, */
    cmf->head = blk->succ;      /* remove block from the list */
    free(blk);                  /* (go to next), and delete block */
  }
  free(cmf);                    /* delete the base structure */
}  /* cmf_delete() */

/*--------------------------------------------------------------------*/

int cmf_add (CMFILTER *cmf, const ITEM *items, ITEM n, RSUPP supp)
{                               /* --- add an item set */
  size_t  size;                 /* size of stored item set */
  CMFBLK  *blk;                 /* current/created memory block */
  ITEMSET *iset;                /* memory to store item set in */

  assert(cmf                    /* check the function arguments */
  &&  (items || (n <= 0)) && (supp >= 0));
  size = sizeof(ITEMSET) +(size_t)(n-1) *sizeof(ITEM);
  size = (size+7) & ~(size_t)7; /* compute stored size (mult. of 8) */
  blk = (CMFBLK*)cmf->head;     /* get the current memory block */
  if ((char*)blk->end +size > (char*)cmf->end) {
    /* if there is not enough space left in current memory block */
    if (size >= (cmf->blksize >> 3)) {
      blk = (CMFBLK*)malloc(sizeof(CMFBLK) +size);
      if (!blk) return -1;      /* separate block for very large set */
      blk->succ = ((CMFBLK*)cmf->head)->succ;
      blk->end  = blk+1;        /* add new block as second in list */
      ((CMFBLK*)cmf->head)->succ = blk; }
    else {                      /* if item set is small */
      blk = (CMFBLK*)malloc(cmf->blksize);
      if (!blk) return -1;      /* create a new memory block */
      blk->succ = cmf->head;    /* (of standard size) and */
      blk->end  = blk+1;        /* add it at the head */
      cmf->head = blk;          /* of the memory block list */
      cmf->end  = (void*)((char*)blk +cmf->blksize);
    }                           /* compute the end of the block */
    cmf->blkcnt += 1;           /* count the created block */
  }
  iset = (ITEMSET*)blk->end;    /* get storage location of item set */
  blk->end = (void*)((char*)blk->end +size);
  iset->supp = supp;            /* store the item set support */
  iset->size = n;               /* and the number of items */
  memcpy(iset->items, items, (size_t)n *sizeof(ITEM));
  cmf->iscnt += 1;              /* copy items and count item set */
  return 0;                     /* return 'ok' */
}  /* cmf_add() */

/*--------------------------------------------------------------------*/

static int cmpsize (const void* a, const void* b, void *data)
{                               /* --- compare item set sizes */
  if (((ITEMSET*)a)->size < ((ITEMSET*)b)->size) return -1;
  if (((ITEMSET*)a)->size > ((ITEMSET*)b)->size) return +1;
  return 0;                     /* return sign of size difference */
}  /* cmpsize() */

/*--------------------------------------------------------------------*/

static int cmpsupp (const void* a, const void* b, void *data)
{                               /* --- compare item set sizes */
  if (((ITEMSET*)a)->supp < ((ITEMSET*)b)->supp) return -1;
  if (((ITEMSET*)a)->supp > ((ITEMSET*)b)->supp) return +1;
  return 0;                     /* return sign of size difference */
}  /* cmpsupp() */

/*--------------------------------------------------------------------*/

static int subset (const ITEMSET *a, const ITEMSET *b)
{                               /* --- check subset relationship */
  ITEM i, k;                    /* loop variables for items */

  i = a->size;                  /* if the first item set is empty, */
  if (--i < 0) return -1;       /* it is a subset of any item set */
  for (k = b->size; --k >= i; ) /* traverse items in destination */
    if ((a->items[i] == b->items[k]) && (--i < 0))
      return -1;                /* if all source items matched, */
  return 0;                     /* success, otherwise, failure */
}  /* subset() */

/*--------------------------------------------------------------------*/

static ITEMSET** getisets (CMFILTER *cmf)
{                               /* --- get array of item sets */
  size_t  i = 0;                /* index of next item set */
  ITEMSET **isets;              /* created item set array */
  ITEMSET *iset;                /* to traverse the item sets */
  CMFBLK  *blk;                 /* to traverse the memory blocks */
  size_t  size;                 /* size of stored item set */
  char    *p;                   /* current location in memory block */

  isets = (ITEMSET**)malloc(cmf->iscnt *sizeof(ITEMSET*));
  if (!isets) return NULL;      /* create an item set array */
  for (blk = cmf->head; blk; blk = blk->succ) {
    /* traverse the list of memory blocks */
    for (p = (char*)(blk+1); p < (char*)blk->end; ) {
      isets[i++] = iset = (ITEMSET*)p;
      size = sizeof(ITEMSET) +(size_t)(iset->size-1) *sizeof(ITEM);
      p += (size+7) & ~(size_t)7;
    }                           /* get the next item set and advance */
  }                             /* the location by its stored size */
  return isets;                 /* return created item set array */
}  /* getisets() */

/*--------------------------------------------------------------------*/

int cmf_report (CMFILTER *cmf, ISREPORT *report)
{                               /* --- report all item sets */
  int      r;                   /* result of reporting function */
  ITEMSET *iset;                /* to traverse the item sets */
  CMFBLK  *blk;                 /* to traverse the memory blocks */
  size_t  size;                 /* size of stored item set */
  char    *p;                   /* current location in memory block */

  for (blk = cmf->head; blk; blk = blk->succ) {
    /* traverse the list of memory blocks */
    for (p = (char*)(blk+1); p < (char*)blk->end; ) {
      iset = (ITEMSET*)p;       /* traverse item sets in block */
      r = isr_iset(report, iset->items, iset->size, iset->supp, 0, 0);
      if (r < 0) return r;      /* report the current item set */
      size = sizeof(ITEMSET) +(size_t)(iset->size-1) *sizeof(ITEM);
      p += (size+7) & ~(size_t)7;
    }                           /* advance the location */
  }                             /* by the item set's stored size */
  return 0;                     /* return 'ok' */
}  /* cmf_report() */

/*--------------------------------------------------------------------*/

int cmf_closed (CMFILTER *cmf, ISREPORT *report)
{                               /* --- report closed item sets */
  int     r;                    /* result of reporting function */
  size_t  i, k;                 /* loop variables for item sets */
  ITEMSET **isets;              /* array of item sets */
  size_t  beg, end;             /* range of item sets with same supp. */

  assert(cmf && report);        /* check the function arguments */
  isets = getisets(cmf);        /* collect the stored item sets */
  if (!isets) return -1;        /* in an array and sort them */
  ptr_qsort(isets, cmf->iscnt, +1, cmpsupp, NULL);
  for (beg = 0; beg < cmf->iscnt; beg = end) {
    for (end = beg+1; end < cmf->iscnt; end++)
      if (isets[end]->supp != isets[beg]->supp)
        break;                  /* find section with same support */
    ptr_qsort(isets+beg, end-beg, -1, cmpsize, NULL);
    for (i = beg; i < end; i++) {
      for (k = beg; k < i; k++) {
        if ((isets[k] != NULL) && subset(isets[i], isets[k])) {
          isets[i] = NULL; break; }
      }                         /* if superset found, delete item set */
      if (!isets[i]) continue;  /* if item set has superset, skip it */
      r = isr_iset(report, isets[i]->items, isets[i]->size,
                           isets[i]->supp, 0, 0);
      if (r < 0) return r;      /* report the item set and */
    }                           /* abort if an error occurred */
  }
  return 0;                     /* return 'ok' */
}  /* cmf_closed() */

/*--------------------------------------------------------------------*/

int cmf_maximal (CMFILTER *cmf, ISREPORT *report)
{                               /* --- report maximal item sets */
  int     r;                    /* result of reporting function */
  size_t  i, k;                 /* loop variables for item sets */
  ITEMSET **isets;              /* array of item sets */

  assert(cmf && report);        /* check the function arguments */
  isets = getisets(cmf);        /* collect the stored item sets */
  if (!isets) return -1;        /* in an array and sort them */
  ptr_qsort(isets, cmf->iscnt, -1, cmpsize, NULL);
  for (i = 0; i < cmf->iscnt; i++) {
    for (k = 0; k < i; k++) {   /* check pairs for subset relation */
      if ((isets[k] != NULL) && subset(isets[i], isets[k])) {
        isets[i] = NULL; break; }
    }                           /* if superset found, delete item set */
    if (!isets[i]) continue;    /* if item set has superset, skip it */
    r = isr_iset(report, isets[i]->items, isets[i]->size,
                         isets[i]->supp, 0, 0);
    if (r < 0) return r;        /* report the item set and */
  }                             /* abort if an error occurred */
  return 0;                     /* return 'ok' */
}  /* cmf_maximal() */
